import pytest

import autoconf as aconf
from autoconf.mock.mock_real import SphProfile


@pytest.fixture(name="geometry_profile_path")
def make_geometry_profile_path():
    return ["autoconf", "mock", "mock_real", "SphProfile"]


def test_path_for_class(geometry_profile_path):
    assert aconf.path_for_class(SphProfile) == geometry_profile_path


@pytest.mark.parametrize(
    "config_dict, paths",
    [
        (
            {
                "autoconf.mock.mock_real.SphProfile": "test",
                "autoconf.mock.mock_real.Other": "toast",
            },
            ["autoconf.mock.mock_real.SphProfile", "autoconf.mock.mock_real.Other"],
        ),
        (
            {"autoconf.mock.mock_real": {"SphProfile": "test", "Other": "toast"}},
            [
                "autoconf.mock.mock_real",
                "autoconf.mock.mock_real.SphProfile",
                "autoconf.mock.mock_real.Other",
            ],
        ),
        (
            {"autoconf": {"mock": {"mock_real": {"SphProfile": "test", "Other": "toast"}}}},
            [
                "autoconf",
                "autoconf.mock",
                "autoconf.mock.mock_real",
                "autoconf.mock.mock_real.SphProfile",
                "autoconf.mock.mock_real.Other",
            ],
        ),
        (
            {"autoconf": {"mock": {"mock_real.SphProfile": "test", "mock_real.Other": "toast"}}},
            [
                "autoconf",
                "autoconf.mock",
                "autoconf.mock.mock_real.SphProfile",
                "autoconf.mock.mock_real.Other",
            ],
        ),
        ({"SphProfile": "test", "Other": "toast"}, ["SphProfile", "Other"]),
        (
            {"mock_real.SphProfile": "test", "mock_real.Other": "toast"},
            ["mock_real.SphProfile", "mock_real.Other"],
        ),
        (
            {"mock_real": {"SphProfile": "test", "Other": "toast"}},
            ["mock_real", "mock_real.SphProfile", "mock_real.Other"],
        ),
    ],
)
def test_paths(config_dict, paths):
    config = aconf.JSONPriorConfig(config_dict)
    assert config.paths == paths


@pytest.mark.parametrize(
    "config_dict",
    [
        {
            "autoconf.mock.mock_real.SphProfile": "test",
            "autoconf.mock.mock_real.Other": "toast",
        },
        {"autoconf.mock.mock_real": {"SphProfile": "test", "Other": "toast"}},
        {"autoconf":{"mock": {"mock_real": {"SphProfile": "test", "Other": "toast"}}}},
        {"autoconf":{"mock": {"mock_real.SphProfile": "test", "mock_real.Other": "toast"}}},
        {"SphProfile": "test", "Other": "toast"},
        {"mock_real": {"SphProfile": "test", "Other": "toast"}},
    ],
)
def test_config_for_path(geometry_profile_path, config_dict):
    config = aconf.JSONPriorConfig(config_dict)
    assert config(geometry_profile_path) == "test"
    assert config(["autoconf", "mock", "mock_real", "Other"]) == "toast"


def test_path_double():
    config = aconf.JSONPriorConfig({"mock_real": {"SphProfile": "test"}})
    assert config(["something", "mock_real", "mock_real", "SphProfile"]) == "test"



from autoconf import conf
from autoconf.mock.mock_real import Redshift


def test_missing_logging_config(
        files_directory
):
    config = conf.Config(
        files_directory / "no_logging",
    )
    config.push(new_path=files_directory / "no_logging")


def test_override_file(config):
    hpc = config["general"]["hpc"]

    assert hpc["hpc_mode"] is False
    assert hpc["default_field"] == "hello"


def test_doesnt_exist(config):
    config.push("/non/existent/path")
    assert config.logging_config["name"] == "config"


def test_logging_config(config, files_directory):
    assert config.logging_config["name"] == "config"

    config.push(files_directory / "default")
    assert config.logging_config["name"] == "default"


def test_push(config, files_directory):
    assert len(config.configs) == 2
    assert config["general"]["hpc"]["hpc_mode"] is False

    config.push(files_directory / "default")

    assert len(config.configs) == 2
    assert config["general"]["hpc"]["hpc_mode"] is True

    config.push(files_directory / "config")

    assert len(config.configs) == 2
    assert config["general"]["hpc"]["hpc_mode"] is False


def test_keep_first(config, files_directory):
    config.push(
        files_directory / "default",
        keep_first=True
    )

    assert config["general"]["hpc"]["hpc_mode"] is False


def test_override_in_directory(config):
    subscript = config["text"]["label"]["subscript"]

    assert subscript["Galaxy"] == "g"
    assert subscript["default_field"] == "label default"


def test_novel_directory(config):
    assert config["default"]["other"]["section"]["key"] == "value"


def test_novel_file(config):
    assert config["default_file"]["section"]["key"] == "file value"


def test_json(config):
    assert config.prior_config.for_class_and_suffix_path(
        Redshift, ["redshift"]
    )["upper_limit"] == 3.0
    assert config.prior_config.for_class_and_suffix_path(
        Redshift, ["rodshift"]
    )["upper_limit"] == 4.0



import json
import os
from pathlib import Path

import pytest

from autoconf.json_prior import generate as g

directory = Path(__file__).parent
package_directory = directory / "code"
module_path = package_directory / "module.py"


@pytest.fixture(
    name="prior_json"
)
def make_prior_json():
    return {
        'MyClass': {
            'simple': {
                'gaussian_limits': {'lower': 0.0, 'upper': 1.0},
                'lower_limit': 0.0,
                'type': 'Uniform',
                'upper_limit': 1.0,
                'width_modifier': {'type': 'Absolute', 'value': 0.2}
            },
            'tup_0': {
                'gaussian_limits': {'lower': 0.0, 'upper': 1.0},
                'lower_limit': 0.0,
                'type': 'Uniform',
                'upper_limit': 1.0,
                'width_modifier': {'type': 'Absolute', 'value': 0.2}
            },
            'tup_1': {
                'gaussian_limits': {'lower': 0.0, 'upper': 1.0},
                'lower_limit': 0.0,
                'type': 'Uniform',
                'upper_limit': 1.0,
                'width_modifier': {'type': 'Absolute', 'value': 0.2}
            }
        }
    }


@pytest.fixture(
    autouse=True
)
def cleanup():
    yield
    try:
        os.remove("priors/module.json")
    except FileNotFoundError:
        pass


def test_generate_for_file(prior_json):
    result = g.for_file(module_path)
    assert result == prior_json


def test_generate(prior_json):
    g.generate(package_directory)

    with open(f"priors/module.json") as f:
        assert json.load(f) == prior_json



import pytest

from autoconf import conf
from autoconf.conf import with_config


@pytest.fixture(
    autouse=True
)
def push_configs(
        files_directory
):
    conf.instance.push(files_directory / "config")
    conf.instance.push(files_directory / "default")


@with_config(
    "general",
    "output",
    "identifier_version",
    value=9
)
def test_with_config():
    assert conf.instance["general"]["output"]["identifier_version"] == 9


def test_config():
    assert conf.instance["general"]["output"]["identifier_version"] == 4



import pathlib

import pytest

from autoconf import conf


@pytest.fixture(
    name="files_directory"
)
def make_files_directory():
    return pathlib.Path(
        __file__
    ).parent / "files"


@pytest.fixture(
    name="config"
)
def make_config(files_directory):
    return conf.Config(
        files_directory / "config",
        files_directory / "default",
    )



from os import path

import pytest

from autoconf.directory_config import NamedConfig
from autoconf.mock.mock_real import EllProfile, EllGaussian

directory = path.dirname(path.realpath(__file__))


class MockClass:
    pass


@pytest.fixture(name="label_config")
def make_label_config():
    return NamedConfig(
        "{}/files/config/label.ini".format(directory)
    )


class TestLabel:
    def test_basic(self, label_config):
        assert label_config["label"]["centre_0"] == "x"
        assert label_config["label"]["redshift"] == "z"

    def test_escaped(self, label_config):
        assert label_config["label"]["gamma"] == r"\gamma"
        assert label_config["label"]["contribution_factor"] == r"\omega0"

    def test_subscript(self, label_config):
        assert label_config["subscript"].family(EllProfile) == "l"

    def test_inheritance(self, label_config):
        assert label_config["subscript"].family(EllGaussian) == "l"

    def test_exception(self, label_config):
        with pytest.raises(KeyError):
            label_config["subscript"].family(MockClass)



import configparser
import os
from abc import abstractmethod, ABC
from pathlib import Path
from autoconf import exc


class AbstractConfig(ABC):
    @abstractmethod
    def _getitem(self, item):
        pass

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.items()[item]
        return self._getitem(item)

    def items(self):
        return [(key, self[key]) for key in self.keys()]

    def __len__(self):
        return len(self.items())

    @abstractmethod
    def keys(self):
        pass

    def family(self, cls):
        for cls in family(cls):
            key = cls.__name__
            try:
                return self[key]
            except (KeyError, configparser.NoOptionError):
                pass
        raise KeyError(
            f"No configuration found for {cls.__name__}"
        )


class SectionConfig(AbstractConfig):
    def __init__(self, path, parser, section):
        self.path = path
        self.section = section
        self.parser = parser

    def keys(self):
        return [item[0] for item in self.parser.items(self.section)]

    def _getitem(self, item):
        try:
            result = self.parser.get(
                self.section,
                item
            )
            if result.lower() == "true":
                return True
            if result.lower() == "false":
                return False
            if result.lower() in ("none", "null"):
                return None
            if result.isdigit():
                return int(result)
            try:
                return float(result)
            except ValueError:
                return result
        except (configparser.NoSectionError, configparser.NoOptionError):
            raise KeyError(
                f"No configuration found for {item} at path {self.path}"
            )


class NamedConfig(AbstractConfig):

    def __init__(self, config_path : str):
        """
        Parses generic config

        Parameters
        ----------
        config_path
            The path to the config file
        """
        self.path = config_path
        self.parser = configparser.ConfigParser()
        self.parser.read(self.path)

    def keys(self):
        return self.parser.sections()

    def _getitem(self, item):
        return SectionConfig(
            self.path,
            self.parser,
            item,
        )


class RecursiveConfig(AbstractConfig):
    def keys(self):
        try:
            return [
                path.split(".")[0]
                for path
                in os.listdir(self.path)
                if all([
                    path != "priors",
                    len(path.split(".")[0]) != 0,
                    os.path.isdir(
                        f"{self.path}/{path}"
                    ) or path.endswith(".ini")
                ])
            ]
        except FileNotFoundError as e:
            raise KeyError(
                f"No configuration found at {self.path}"
            ) from e

    def __init__(self, path):
        self.path = Path(path)

    def __eq__(self, other):
        return str(self) == str(other)

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.path}>"

    def _getitem(self, item):
        item_path = self.path / f"{item}"
        file_path = f"{item_path}.ini"
        if os.path.isfile(file_path):
            return NamedConfig(
                file_path
            )
        if os.path.isdir(item_path):
            return RecursiveConfig(
                item_path
            )
        raise KeyError(
            f"No configuration found for {item} at path {self.path}"
        )


class PriorConfigWrapper:
    def __init__(self, prior_configs):
        self.prior_configs = prior_configs

    def for_class_and_suffix_path(self, cls, path):
        for config in self.prior_configs:
            try:
                return config.for_class_and_suffix_path(
                    cls, path
                )
            except KeyError:
                pass
        directories = ' '.join(
            str(config.directory) for config in self.prior_configs
        )

        print(

        )

        raise exc.ConfigException(
            f"No prior config found for class: \n\n"
            f"{cls.__name__} \n\n"
            f"For parameter name and path: \n\n "
            f"{'.'.join(path)} \n\n "
            f"In any of the following directories:\n\n"
            f"{directories}\n\n"
            f"The following readthedocs page explains prior configuration files in PyAutoFit and will help you fix "
            f"the error https://pyautofit.readthedocs.io/en/latest/general/adding_a_model_component.html"
        )


class ConfigWrapper(AbstractConfig):
    def __init__(self, configs):
        self.configs = configs

    @property
    def paths(self):
        return [
            config.path
            for config
            in self.configs
        ]

    def __applicable(self, item):
        __applicable = list()
        for config in self.configs:
            try:
                __applicable.append(config[item])
            except KeyError:
                pass
        return __applicable

    def items(self):
        item_dict = {}
        for config in reversed(
                self.configs
        ):
            for key, value in config.items():
                item_dict[key] = value
        return list(item_dict.items())

    def keys(self):
        keys = set()
        for config in self.configs:
            keys.update(config.keys())
        return list(keys)

    def _getitem(self, item):
        configs = self.__applicable(item)
        if len(configs) == 0:
            paths = '\n'.join(map(str, self.paths))
            raise KeyError(
                f"No configuration for {item} in {paths}"
            )
        for config in configs:
            if not isinstance(
                    config, AbstractConfig
            ):
                return config
        return ConfigWrapper(configs)


def family(current_class):
    yield current_class
    for next_class in current_class.__bases__:
        for val in family(next_class):
            yield val


import logging
import logging.config
import os
import shutil
from functools import wraps
from pathlib import Path
from typing import Optional

import yaml

from autoconf.directory_config import RecursiveConfig, PriorConfigWrapper, AbstractConfig, family
from autoconf.json_prior.config import JSONPriorConfig

logger = logging.getLogger(__name__)

LOGGING_CONFIG_FILE = "logging.yaml"


def get_matplotlib_backend():
    try:
        return instance["visualize"]["general"]["general"]["backend"]
    except KeyError:
        return "default"


class DictWrapper:
    def __init__(self, paths):
        self._dict = dict()
        self.paths = paths

    def __contains__(self, item):
        return item in self._dict

    def items(self):
        return self._dict.items()

    def __setitem__(self, key, value):
        if isinstance(key, str):
            key = key.lower()
        self._dict[key] = value

    def __getitem__(self, key):
        if isinstance(key, str):
            key = key.lower()
        try:
            return self._dict[key]
        except KeyError:
            raise KeyError(
                f"key {key} not found in paths {self.paths_string}"
            )

    @property
    def paths_string(self):
        return "\n".join(
            map(str, self.paths)
        )

    def __repr__(self):
        return repr(self._dict)

    def family(self, cls):
        for item in family(cls):
            try:
                return self[item]
            except KeyError:
                pass
        raise KeyError(
            f"config for {cls} or its parents not found in paths {self.paths_string}"
        )


class Config:
    def __init__(self, *config_paths, output_path="output"):
        """
        Singleton to manage configuration.

        Configuration is loaded using the __getitem__ syntax where the key entered
        can refer to a directory, file, section or item.

        Configuration is first attempted to be loaded from the directory indicated by the first
        config_path. If no configuration is found the second directory is searched and so on.
        This allows a default configuration to be defined with additional configuration overriding
        it.

        Parameters
        ----------
        config_paths
            Indicate directories where configuration is defined, in the order of priority with
            configuration in the first config_path overriding configuration in later config
            paths
        output_path
            The path where data should be saved.
        """
        self._configs = list()
        self._dict = DictWrapper(
            self.paths
        )

        self.configs = list(map(
            RecursiveConfig,
            config_paths
        ))

        self.output_path = output_path

    def configure_logging(self):
        """
        Set the most up to date logging configuration
        """
        logging_config = self.logging_config
        if logging_config is not None:
            logging.config.dictConfig(
                logging_config
            )

    @property
    def logging_config(self) -> Optional[dict]:
        """
        Loading logging configuration from a YAML file
        from the most recently added config directory
        for which it exists.
        """
        for config in self.configs:
            path = config.path
            try:
                if LOGGING_CONFIG_FILE in os.listdir(
                        config.path
                ):
                    with open(
                            path / LOGGING_CONFIG_FILE
                    ) as f:
                        return yaml.safe_load(f)
            except FileNotFoundError:
                logger.warning(
                    f"No configuration found at path {config.path}"
                )
        return None

    @property
    def configs(self):
        return self._configs

    @configs.setter
    def configs(self, configs):
        self._configs = configs

        def recurse_config(
                config,
                d
        ):
            try:
                for key, value in config.items():
                    if isinstance(
                            value,
                            AbstractConfig
                    ):
                        if key not in d:
                            d[key] = DictWrapper(
                                self.paths
                            )
                        recurse_config(
                            value,
                            d=d[key]
                        )
                    else:
                        d[key] = value
            except KeyError as e:
                logger.debug(e)

        for config_ in reversed(configs):
            recurse_config(config_, self._dict)

    def __getitem__(self, item):
        return self._dict[item]

    @property
    def paths(self):
        return [
            config.path
            for config
            in self._configs
        ]

    @property
    def prior_config(self) -> PriorConfigWrapper:
        """
        Configuration for priors. This indicates, for example, the mean and the width of priors
        for the attributes of given classes.
        """
        return PriorConfigWrapper([
            JSONPriorConfig.from_directory(
                path / "priors"
            )
            for path in self.paths
        ])

    def push(
            self,
            new_path: str,
            output_path: Optional[str] = None,
            keep_first: bool = False
    ):
        """
        Push a new configuration path. This overrides the existing config
        paths, with existing configs being used as a backup when a value
        cannot be found in an overriding config.

        Parameters
        ----------
        new_path
            A path to config directory
        output_path
            The path at which data should be output. If this is None then it remains
            unchanged
        keep_first
            If True the current priority configuration mains such.
        """
        logger.warning(
            f"Pushing new config with path {new_path}"
        )

        self.output_path = output_path or self.output_path

        if self.configs[0] == new_path or (
                keep_first and len(self.configs) > 1 and self.configs[1] == new_path
        ):
            return
        new_config = RecursiveConfig(
            new_path
        )

        configs = list(filter(
            lambda config: config != new_config,
            self.configs
        ))
        if keep_first:
            self.configs = configs[:1] + [new_config] + configs[1:]
        else:
            self.configs = [new_config] + configs

        self.configure_logging()

    def register(self, file: str):
        """
        Add defaults for a given project

        Parameters
        ----------
        file
            The path to the project's __init__
        """
        self.push(
            Path(file).parent / "config",
            keep_first=True
        )


current_directory = Path(os.getcwd())

default = Config(
    current_directory / "config",
    current_directory / "output/"
)

instance = default


def output_path_for_test(
        temporary_path="temp",
        remove=True
):
    """
    Temporarily change the output path for the scope of a function
    (e.g. a test). Remove the files after the test has completed
    execution.

    Parameters
    ----------
    temporary_path
        The path to temporarily output files to
    remove
        Should the path be removed?

    Returns
    -------
    The original function, decorated
    """

    def remove_():
        if remove:
            shutil.rmtree(
                temporary_path,
                ignore_errors=True
            )

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            remove_()
            original_path = instance.output_path
            instance.output_path = temporary_path

            result = func(*args, **kwargs)

            remove_()
            instance.output_path = original_path

            return result

        return wrapper

    return decorator


def with_config(*path: str, value):
    """
    Create a decorator that swaps a value in configuration
    defined by path for the scope of a test.

    Parameters
    ----------
    path
        A path through config. e.g. "general", "output", "identifier_version"
    value
        The value to temporarily set for the config field

    Returns
    -------
    A decorator
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            config = instance
            for string in path[:-1]:
                config = config[string]

            original_value = config[path[-1]]
            config[path[-1]] = value

            result = func(*args, **kwargs)

            config[path[-1]] = original_value

            return result

        return wrapper

    return decorator



from . import exc
from .conf import Config
from .conf import instance
from .json_prior.config import default_prior
from .json_prior.config import make_config_for_class
from .json_prior.config import path_for_class
from .json_prior.config import JSONPriorConfig

__version__ = '2021.8.12.1'
